//
// A final version of the 'money' class
//
// Includes all the features from Chapter 27
// in a single class, plus a few additional 
// features to make the code more practical 
// and efficient. 
//
// NOTE:
//
// 1. 'Large' currency units are stored in a long integer, 
//    not an int. 
//
// 2. Explicit use of the 'this' pointer is minimised. 
//
// 3. A check for 'valid money' is carried out in all member 
//    functions. 
//
// 4. The prefix and postfix operators are 'shared' 
//    (This is for illustration only - ++ makes little real 
//    sense for this class).
//
// 5. There is a 'currency' field: note how this is used
//    (You can probably think of various ways to extend this feature)
//
// 6. In operators, R_OP = right operand, L_OP = left operand
// 
// Source is: MONEY.CPP
//
// MJP 7 March, 1997
//

#include <assert.h>
#include <iostream.h>

class cMoney
   {
   public:
      // Construction
      cMoney(const char = '$', const int = 0, const int = 0);
      
      // Arithmetic operators
             cMoney operator*(const int) const;          // Mult R_OP * int
      friend cMoney operator*(const int, const cMoney&); // Mult int * R_OP
             cMoney operator-(const cMoney&) const;      // Binary version
             cMoney operator-();                         // Unary version
             cMoney operator+(const cMoney&) const;      // Binary version
             cMoney operator++();                        // Prefix
             cMoney operator++(int);                     // Postfix
      
      // Logical operators
      int operator>(const cMoney&) const;
      int operator<(const cMoney&) const;
      int operator==(const cMoney&) const;
      
      // Stream insertion and stream extraction
      friend ostream& operator<<(ostream&, cMoney);
      friend istream& operator>>(istream&, cMoney&); 
      
   protected:
      // Some binary operations can only be carried out
      // if the currency is the same (or if a conversion
      // is available - but this isn't implemented here)
      int _Is_Same_Currency(const cMoney&) const;
      
      // A useful member function used to check
      // that '_Small' value is valid.        
      int _Is_Valid() const;   

   private:
      // NOTE: We assume there are 100 '_Small' currency units (e.g. cents) to
      // a single '_Large' currency unit (e.g. dollars).  This could be altered
      // quite easily
      long _Large; // Allow for larger values... 
      int  _Small;  
      char _Currency;
   };   
  
// Constructor  
cMoney::cMoney(const char CURR, const int PO, const int PE)
   {
   _Large = PO;
   _Small = PE; 
   _Currency = CURR;
   }   

// Overload * for cMoney class plus int
cMoney cMoney::operator*(const int MULTIPLIER) const
   {
   // Precondition
   assert(_Is_Valid());
   
   const cMoney L_OP = *this; // Left operand
   cMoney Product;
   
   Product._Large = L_OP._Large * MULTIPLIER;
   Product._Small = L_OP._Small * MULTIPLIER;

   while (Product._Small > 99)
      {
      Product._Small-=100;
      Product._Large+=1;
      }     
  
   // Postcondition
   assert(_Is_Valid());
   assert(Product._Is_Valid());  
   
   return Product;
   }

// Overload * for int plus cMoney class (again)
cMoney operator*(const int MULTIPLIER, const cMoney& R_OP)
   {
   // Precondition
   assert(R_OP._Is_Valid());

   cMoney Product;
   
   Product._Large = R_OP._Large * MULTIPLIER;
   Product._Small = R_OP._Small * MULTIPLIER;
   
   while (Product._Small > 99)
      {
      Product._Small-=100;
      Product._Large+=1;
      }     

   // Postcondition
   assert(Product._Is_Valid());  

   return Product;
   }
  

// Overload binary - relative to cMoney class
cMoney cMoney::operator-(const cMoney& R_OP) const
   {
   // Precondition
   assert(_Is_Valid());
   assert(R_OP._Is_Valid());
   assert(_Is_Same_Currency(R_OP)); 

   const cMoney L_OP = *this; // Left operand
   cMoney Diff;
   
   Diff._Large = L_OP._Large - R_OP._Large;
   Diff._Small = L_OP._Small - R_OP._Small;
   
   if (Diff._Small < 0)
      {
      Diff._Small+=100;
      Diff._Large-=1;
      }     
   
   // Postcondition
   assert(_Is_Valid());

   return Diff;
   }


// Overload unary - relative to cMoney class
cMoney cMoney::operator-()
   {
   // Precondition
   assert(_Is_Valid());

   _Large = -_Large;
   
   // Postcondition
   assert(_Is_Valid());

   return *this;
   }


// Overload binary + relative to cMoney class
cMoney cMoney::operator+(const cMoney& R_OP) const
   {
   // Precondition
   assert(_Is_Valid());
   assert(R_OP._Is_Valid());
   assert(_Is_Same_Currency(R_OP)); 

   const cMoney L_OP = *this; // Left operand
   cMoney Sum;
   
   Sum._Large = L_OP._Large + R_OP._Large;
   Sum._Small = L_OP._Small + R_OP._Small;
   
   while (Sum._Small > 99)
      {
      Sum._Small-=100;
      Sum._Large+=1;
      }     
   
   // Postcondition
   assert(Sum._Is_Valid());

   return Sum;
   }


// Overload unary prefix ++ relative to cMoney class
cMoney cMoney::operator++()
   {
   // Precondition
   assert(_Is_Valid());

   ++_Large;
   ++_Small;

   if (_Small > 99)
      {
      _Small-=100;
      _Large+=1;
      }     

   // Postcondition
   assert(_Is_Valid());

   return *this;
   }


// Overload unary postfix ++ relative to cMoney class
cMoney cMoney::operator++(int /* DUMMY PARAMETER */)
   {
   // Precondition
   assert(_Is_Valid());

   cMoney Return_value = *this;
   
   // NB - one solution avoiding duplicate code.
   operator++();
    
   // Postcondition
   assert(_Is_Valid());

   return Return_value;
   }
   

int cMoney::operator>(const cMoney& R_OP) const
   {
   // Precondition
   assert(_Is_Valid());
   assert(R_OP._Is_Valid());
   assert(_Is_Same_Currency(R_OP)); 

   if (((_Large > R_OP._Large)
      || ((_Large == R_OP._Large) 
      && (_Small > R_OP._Small))))
      {
      return 1;
      }
   else
      {
      return 0;
      }   
   }

ostream& operator<<(ostream& Stream, cMoney R_OP)
   {
   // Precondition
   assert(R_OP._Is_Valid());

   if (R_OP._Large < 0)
      {
      Stream << "-" << R_OP._Currency << -R_OP._Large;
      }
   else
      {
      Stream << R_OP._Currency << R_OP._Large;
      }   
   
   if (R_OP._Small < 10)
      {
      Stream << ".0" << R_OP._Small;
      }
   else
      {
      Stream << "." << R_OP._Small;
      } 
      
   return Stream;      
   }

   
istream& operator>>(istream& stream, cMoney& R_OP)
   {
   // Precondition
   assert(R_OP._Is_Valid());

   cout << "Please enter cash amount \n";
   cout << R_OP._Currency << " : ";
   stream >> R_OP._Large;
   cout << ". : ";
   stream >> R_OP._Small;
   cout << R_OP << endl;

   return stream;
   }           


int cMoney::_Is_Valid() const
   {
   // Simple check for validity
   return ((_Small <= 99) && (_Small >= 0));
   }

int cMoney::_Is_Same_Currency(const cMoney& R_OP) const
   {
   // Precondition
   assert(_Is_Valid());

   // Check currency of two R_OPs is same
   return (_Currency == R_OP._Currency);
   }

int main()
   {
   cMoney Cash1('',1,01), Cash2('',2,00), Cash3(''), Cash4('');

   cin >> Cash4;   
   
   cout << endl;
 
   Cash3 = Cash1; 
   cout << "Cash3 = " << Cash3 << endl;
   cout << "Cash2 = " << Cash2 << endl;
   
   if (Cash3++ > Cash2)
      {
      cout << "Cash3++ > Cash2\n";
      }
   else
      {
      cout << "Cash3++ <= Cash2\n";
      }            
      
   cout << "Cash3 = " << Cash3 << endl;
   cout << "Cash2 = " << Cash2 << endl << endl;

   Cash3 = Cash1; 
    
   cout << "Cash3 = " << Cash3 << endl;
   cout << "Cash2 = " << Cash2 << endl << endl;

   if (++Cash3 > Cash2)
      {
      cout << "++Cash3 > Cash2\n";
      }
   else
      {
      cout << "++Cash3 <= Cash2\n";
      }   

   cout << "Cash3 = " << Cash3 << endl;
   cout << "Cash2 = " << Cash2 << endl << endl;

   Cash3 = Cash1;  
   cout << "Cash3 = " << Cash3 << endl;
   cout << "-Cash3 = " << -Cash3 << endl;
   
   return 0;
   }
